/*
 * hostapd / main()
 * Copyright (c) 2002-2017, Jouni Malinen <j@w1.fi>
 *
 * This software may be distributed under the terms of the BSD license.
 * See README for more details.
 */

#include "utils/includes.h"
#ifndef CONFIG_NATIVE_WINDOWS
#include <syslog.h>
#include <grp.h>
#endif /* CONFIG_NATIVE_WINDOWS */

#include "utils/common.h"
#include "utils/eloop.h"
#include "utils/uuid.h"
#include "crypto/random.h"
#include "crypto/tls.h"
#include "common/version.h"
#include "drivers/driver.h"
#include "eap_server/eap.h"
#include "eap_server/tncs.h"
#include "ap/hostapd.h"
#include "ap/ap_config.h"
#include "ap/ap_drv_ops.h"
#include "ap/dpp_hostapd.h"
#include "fst/fst.h"
#include "config_file.h"
#include "eap_register.h"
#include "ctrl_iface.h"
#include "systemd/sd-daemon.h"
#include "radius/radius_das.h"


void wired_acl_signal(void *eloop_ctx, void *timeout_ctx);

struct hapd_global {
	void **drv_priv;
	size_t drv_count;
};

static struct hapd_global global;


#ifndef CONFIG_NO_HOSTAPD_LOGGER
static void hostapd_logger_cb(void *ctx, const u8 *addr, unsigned int module,
			      int level, const char *txt, size_t len)
{
	struct hostapd_data *hapd = ctx;
	char *format, *module_str;
	int maxlen;
	int conf_syslog_level, conf_stdout_level;
	unsigned int conf_syslog, conf_stdout;

	maxlen = len + 100;
	format = os_malloc(maxlen);
	if (!format)
		return;

	if (hapd && hapd->conf) {
		conf_syslog_level = hapd->conf->logger_syslog_level;
		conf_stdout_level = hapd->conf->logger_stdout_level;
		conf_syslog = hapd->conf->logger_syslog;
		conf_stdout = hapd->conf->logger_stdout;
	} else {
		conf_syslog_level = conf_stdout_level = 0;
		conf_syslog = conf_stdout = (unsigned int) -1;
	}

	switch (module) {
	case HOSTAPD_MODULE_IEEE80211:
		module_str = "IEEE 802.11";
		break;
	case HOSTAPD_MODULE_IEEE8021X:
		module_str = "IEEE 802.1X";
		break;
	case HOSTAPD_MODULE_RADIUS:
		module_str = "RADIUS";
		break;
	case HOSTAPD_MODULE_WPA:
		module_str = "WPA";
		break;
	case HOSTAPD_MODULE_DRIVER:
		module_str = "DRIVER";
		break;
	case HOSTAPD_MODULE_IAPP:
		module_str = "IAPP";
		break;
	case HOSTAPD_MODULE_MLME:
		module_str = "MLME";
		break;
	default:
		module_str = NULL;
		break;
	}

	if (hapd && hapd->conf && addr)
		os_snprintf(format, maxlen, "%s: STA " MACSTR "%s%s: %s",
			    hapd->conf->iface, MAC2STR(addr),
			    module_str ? " " : "", module_str ? module_str : "",
			    txt);
	else if (hapd && hapd->conf)
		os_snprintf(format, maxlen, "%s:%s%s %s",
			    hapd->conf->iface, module_str ? " " : "",
			    module_str ? module_str : "", txt);
	else if (addr)
		os_snprintf(format, maxlen, "STA " MACSTR "%s%s: %s",
			    MAC2STR(addr), module_str ? " " : "",
			    module_str ? module_str : "", txt);
	else
		os_snprintf(format, maxlen, "%s%s%s",
			    module_str ? module_str : "",
			    module_str ? ": " : "", txt);

#ifdef CONFIG_DEBUG_SYSLOG
	if (wpa_debug_syslog)
		conf_stdout = 0;
#endif /* CONFIG_DEBUG_SYSLOG */
	if ((conf_stdout & module) && level >= conf_stdout_level) {
		wpa_debug_print_timestamp();
		wpa_printf(MSG_INFO, "%s", format);
	}

#ifndef CONFIG_NATIVE_WINDOWS
	if ((conf_syslog & module) && level >= conf_syslog_level) {
		int priority;
		switch (level) {
		case HOSTAPD_LEVEL_DEBUG_VERBOSE:
		case HOSTAPD_LEVEL_DEBUG:
			priority = LOG_DEBUG;
			break;
		case HOSTAPD_LEVEL_INFO:
			priority = LOG_INFO;
			break;
		case HOSTAPD_LEVEL_NOTICE:
			priority = LOG_NOTICE;
			break;
		case HOSTAPD_LEVEL_WARNING:
			priority = LOG_WARNING;
			break;
		default:
			priority = LOG_INFO;
			break;
		}
		syslog(priority, "%s", format);
	}
#endif /* CONFIG_NATIVE_WINDOWS */

	os_free(format);
}
#endif /* CONFIG_NO_HOSTAPD_LOGGER */


/**
 * hostapd_driver_init - Preparate driver interface
 */
static int hostapd_driver_init(struct hostapd_iface *iface)
{
	struct wpa_init_params params;
	size_t i;
	struct hostapd_data *hapd = iface->bss[0];
	struct hostapd_bss_config *conf = hapd->conf;
	u8 *b = conf->bssid;
	struct wpa_driver_capa capa;

	if (hapd->driver == NULL || hapd->driver->hapd_init == NULL) {
		wpa_printf(MSG_ERROR, "No hostapd driver wrapper available");
		return -1;
	}

	/* Initialize the driver interface */
	if (!(b[0] | b[1] | b[2] | b[3] | b[4] | b[5]))
		b = NULL;

	os_memset(&params, 0, sizeof(params));
	for (i = 0; wpa_drivers[i]; i++) {
		if (wpa_drivers[i] != hapd->driver)
			continue;

		if (global.drv_priv[i] == NULL &&
		    wpa_drivers[i]->global_init) {
			global.drv_priv[i] =
				wpa_drivers[i]->global_init(iface->interfaces);
			if (global.drv_priv[i] == NULL) {
				wpa_printf(MSG_ERROR, "Failed to initialize "
					   "driver '%s'",
					   wpa_drivers[i]->name);
				return -1;
			}
		}

		params.global_priv = global.drv_priv[i];
		break;
	}
	params.bssid = b;
	params.ifname = hapd->conf->iface;
	params.driver_params = hapd->iconf->driver_params;
	params.use_pae_group_addr = hapd->conf->use_pae_group_addr;

	params.num_bridge = hapd->iface->num_bss;
	params.bridge = os_calloc(hapd->iface->num_bss, sizeof(char *));
	if (params.bridge == NULL)
		return -1;
	for (i = 0; i < hapd->iface->num_bss; i++) {
		struct hostapd_data *bss = hapd->iface->bss[i];
		if (bss->conf->bridge[0])
			params.bridge[i] = bss->conf->bridge;
	}

	params.own_addr = hapd->own_addr;

	hapd->drv_priv = hapd->driver->hapd_init(hapd, &params);
	os_free(params.bridge);
	if (hapd->drv_priv == NULL) {
		wpa_printf(MSG_ERROR, "%s driver initialization failed.",
			   hapd->driver->name);
		hapd->driver = NULL;
		return -1;
	}

	if (hapd->driver && hapd->drv_priv)
		hapd->driver->ifup(hapd->drv_priv);

	if (hapd->driver->get_capa &&
	    hapd->driver->get_capa(hapd->drv_priv, &capa) == 0) {
		struct wowlan_triggers *triggs;

		iface->drv_flags = capa.flags;
		iface->smps_modes = capa.smps_modes;
		iface->probe_resp_offloads = capa.probe_resp_offloads;
		/*
		 * Use default extended capa values from per-radio information
		 */
		iface->extended_capa = capa.extended_capa;
		iface->extended_capa_mask = capa.extended_capa_mask;
		iface->extended_capa_len = capa.extended_capa_len;
		iface->drv_max_acl_mac_addrs = capa.max_acl_mac_addrs;

		/*
		 * Override extended capa with per-interface type (AP), if
		 * available from the driver.
		 */
		hostapd_get_ext_capa(iface);

		triggs = wpa_get_wowlan_triggers(conf->wowlan_triggers, &capa);
		if (triggs && hapd->driver->set_wowlan) {
			if (hapd->driver->set_wowlan(hapd->drv_priv, triggs))
				wpa_printf(MSG_ERROR, "set_wowlan failed");
		}
		os_free(triggs);
	}

	return 0;
}


/**
 * hostapd_interface_init - Read configuration file and init BSS data
 *
 * This function is used to parse configuration file for a full interface (one
 * or more BSSes sharing the same radio) and allocate memory for the BSS
 * interfaces. No actiual driver operations are started.
 */
static struct hostapd_iface *
hostapd_interface_init(struct hapd_interfaces *interfaces, const char *if_name,
		       const char *config_fname, int debug)
{
	struct hostapd_iface *iface;
	int k;

	wpa_printf(MSG_ERROR, "Configuration file: %s", config_fname);
	iface = hostapd_init(interfaces, config_fname);
	if (!iface)
		return NULL;

	if (if_name) {
		os_strlcpy(iface->conf->bss[0]->iface, if_name,
			   sizeof(iface->conf->bss[0]->iface));
	}

	iface->interfaces = interfaces;

	for (k = 0; k < debug; k++) {
		if (iface->bss[0]->conf->logger_stdout_level > 0)
			iface->bss[0]->conf->logger_stdout_level--;
	}

	if (iface->conf->bss[0]->iface[0] == '\0' &&
	    !hostapd_drv_none(iface->bss[0])) {
		wpa_printf(MSG_ERROR,
			   "Interface name not specified in %s, nor by '-i' parameter",
			   config_fname);
		hostapd_interface_deinit_free(iface);
		return NULL;
	}

	return iface;
}

/**
 * handle_term - SIGINT and SIGTERM handler to terminate hostapd process
 */
static void handle_term(int sig, void *signal_ctx)
{
	wpa_printf(MSG_DEBUG, "Signal %d received - terminating", sig);
	eloop_terminate();
}


#ifndef CONFIG_NATIVE_WINDOWS

/**
 * handle_reload - SIGHUP handler to reload configuration
 */
static void handle_reload(int sig, void *signal_ctx)
{
	struct hostapd_config *conf = NULL;
	struct hapd_interfaces *interfaces = signal_ctx;
	struct hostapd_iface *iface;
	struct hostapd_iface **tmp;
	struct hostapd_iface *new_iface = NULL, *hapd_iface = NULL;
	struct hostapd_data *hapd;
	size_t i, j, k;
	int configured = 0;
	char **if_names = NULL;
	size_t if_names_size = 0;
	char **if_mab_names = NULL;
	size_t if_mab_names_size = 0;
	char **if_pvlan_names = NULL;
	size_t if_pvlan_names_size = 0;
	char **if_voice_names = NULL;
	size_t if_voice_names_size = 0;
	char **delete_if_names = NULL;
	size_t delete_if_names_size = 0;
	char config_filename[256];
	int vlan = 0;

	wpa_printf(MSG_DEBUG, "Signal %d received - reloading configuration",
		   sig);
	if (interfaces && (interfaces->count > 0))
		/* we only want one cl-acltool call at the end */
		interfaces->start_in_progress = TRUE;
	else
		goto fail;

	os_snprintf(config_filename, sizeof(config_filename), "%s",
			interfaces->iface[0]->config_fname);
	/* show what is in the configuration file */
	/* just interfaces */
	if (hostapd_get_interface_names_config(&if_names, &if_names_size,
					       HOSTAPD_BASE_IFACE_CONFIG,
					       config_filename)) {
		wpa_printf(MSG_DEBUG, "Could not read interfaces from config file %s",
				config_filename);
		goto fail;
	}
	/* show what is currently running */
	/* INTERFACE DELETE: */
	for (i = 0; i < interfaces->count; i++) {
		iface = interfaces->iface[i];
		configured = 0;
		hapd = NULL;
		for (j = 0; j < iface->num_bss; j++) {
			hapd = iface->bss[j];
			wpa_printf(MSG_DEBUG, "checking running ifname=%s",
					hapd->conf->iface);
			for (k = 0; k < if_names_size; k++) {
				if (os_strcmp(hapd->conf->iface, if_names[k]) == 0) {
					wpa_printf(MSG_DEBUG, "      is configured");
					configured = 1;
				        break;
				}
			}
			if (configured)
				break;
		}
		if (!configured && hapd && interfaces) {
			wpa_printf(MSG_DEBUG, "       is not configured should be deleted");
			if (hostapd_add_interface_name(&delete_if_names, &delete_if_names_size,
						hapd->conf->iface))
				goto fail;
		}
	}
	for (i = 0; i < delete_if_names_size; i++) {
		hapd = hostapd_get_iface(interfaces, delete_if_names[i]);
		if (hapd->driver && hapd->driver->ifdown && hapd->drv_priv)
			hapd->driver->ifdown(hapd->drv_priv);
		if (hapd->driver && hapd->driver->ifup && hapd->drv_priv)
			hapd->driver->ifup(hapd->drv_priv);
		if (hostapd_disable_iface(hapd->iface))
			wpa_printf(MSG_DEBUG, "Could not disable iface");
		if (hostapd_remove_iface(interfaces, hapd->conf->iface))
			wpa_printf(MSG_DEBUG, "Could not remove iface");
		hostapd_handle_trad_br_removal(delete_if_names[i]);
	}

	if (hostapd_ifnames_free(&delete_if_names, &delete_if_names_size))
		goto fail;

	if (hostapd_get_interface_names_config(&if_mab_names, &if_mab_names_size,
					       HOSTAPD_MAB_IFACE_CONFIG,
					       config_filename)) {
		wpa_printf(MSG_DEBUG, "Could not read MAB interfaces from config file %s",
				config_filename);
		goto fail;
	}
	if (hostapd_get_interface_names_config(&if_pvlan_names, &if_pvlan_names_size,
				               HOSTAPD_PVLAN_IFACE_CONFIG,
					       config_filename)) {
		wpa_printf(MSG_DEBUG, "Could not read PVLAN interfaces from config file %s",
				config_filename);
		goto fail;
	}
	if (hostapd_get_interface_names_config(&if_voice_names, &if_voice_names_size,
				               HOSTAPD_VOICE_IFACE_CONFIG,
					       config_filename)) {
		wpa_printf(MSG_DEBUG, "Could not read Voice interfaces from config file %s",
				config_filename);
		goto fail;
	}

	/* INTERFACE ADD: loop over configured names and intialize any that are not running */
	for (i = 0; i < if_names_size; i++) {
		wpa_printf(MSG_DEBUG, "checking configured ifname=%s", if_names[i]);

		if (hostapd_ifname_in_use(interfaces, if_names[i])) {
			/* do not reinitialize the interface */
			hapd = hostapd_get_iface(interfaces, if_names[i]);
			hostapd_interface_reload(hapd);

			continue;
		}
		/* we have a new interface, initialize it */
		wpa_printf(MSG_DEBUG, "     is NEW, need to initialize");

		hapd_iface = hostapd_interface_init(interfaces,
				if_names[i], config_filename, 0);
		if (!hapd_iface)
			break;
		/* for new interface, make sure MAB is set correctly and init_mab
		*  will get called later in wired_init_mab() */
		for (j = 0; j < if_mab_names_size; j++) {
			if (hapd_iface->bss[0] &&
					os_strcmp(if_names[i], if_mab_names[j]) == 0) {
				hapd_iface->bss[0]->conf->mab_enabled = TRUE;
				break;
			}
		}
		/* some interfaces need to be tagged for to use parking vlans */
		for (j = 0; j < if_pvlan_names_size; j++) {
			if (hapd_iface->bss[0] &&
					os_strcmp(if_names[i], if_pvlan_names[j]) == 0) {
				hapd_iface->bss[0]->conf->pvlan_enabled = TRUE;
				break;
			}
		}
		/* some interfaces need to be tagged as voice enabled */
		for (j = 0; j < if_voice_names_size; j++) {
			if (hapd_iface->bss[0] &&
					os_strcmp(if_names[i], if_voice_names[j]) == 0) {
				hapd_iface->bss[0]->conf->voice_enabled = TRUE;
				/* check for a VLAN (usually for voice) */
				vlan = hostapd_get_interface_vlan(if_voice_names[j]);
				if (vlan)
					hapd_iface->bss[0]->conf->voice_vlan_id = vlan;
				break;
			}
		}
		tmp = os_realloc_array(interfaces->iface,
				interfaces->count + 1,
				sizeof(struct hostapd_iface *));
		if (!tmp) {
			hostapd_interface_deinit_free(hapd_iface);
			break;
		}
		interfaces->iface = tmp;
		interfaces->iface[interfaces->count++] = hapd_iface;
		if (interfaces->driver_init(hapd_iface))
			goto fail;

		if (hostapd_setup_interface(hapd_iface)) {
			hostapd_deinit_driver(
					hapd_iface->bss[0]->driver,
					hapd_iface->bss[0]->drv_priv,
					hapd_iface);
			goto fail;
		}

	}

	interfaces->start_in_progress = FALSE;

	wired_acl_signal(NULL, NULL);

	if (hostapd_ifnames_free(&if_names, &if_names_size))
		goto fail;
	if (hostapd_ifnames_free(&if_mab_names, &if_mab_names_size))
		goto fail;
	if (hostapd_ifnames_free(&if_pvlan_names, &if_pvlan_names_size))
		goto fail;
	if (hostapd_ifnames_free(&if_voice_names, &if_voice_names_size))
		return;

	return;

fail:
	if (conf)
		hostapd_config_free(conf);
	if (hapd_iface) {
		if (hapd_iface->bss) {
			for (i = 0; i < hapd_iface->num_bss; i++) {
				hapd = hapd_iface->bss[i];
				if (!hapd)
					continue;
				if (hapd_iface->interfaces &&
				    hapd_iface->interfaces->ctrl_iface_deinit)
					hapd_iface->interfaces->
						ctrl_iface_deinit(hapd);
				wpa_printf(MSG_DEBUG, "%s: free hapd %p (%s)",
					   __func__, hapd_iface->bss[i],
					   hapd->conf->iface);
				hostapd_cleanup(hapd);
				os_free(hapd);
				hapd_iface->bss[i] = NULL;
			}
			os_free(hapd_iface->bss);
			hapd_iface->bss = NULL;
		}
		if (new_iface) {
			interfaces->count--;
			interfaces->iface[interfaces->count] = NULL;
		}
		hostapd_cleanup_iface(hapd_iface);
	}
	hostapd_ifnames_free(&if_names, &if_names_size);
	hostapd_ifnames_free(&if_mab_names, &if_mab_names_size);
	hostapd_ifnames_free(&if_pvlan_names, &if_pvlan_names_size);
	hostapd_ifnames_free(&if_voice_names, &if_voice_names_size);

	return;
}


static void handle_dump_state(int sig, void *signal_ctx)
{
	/* Not used anymore - ignore signal */
}
#endif /* CONFIG_NATIVE_WINDOWS */


static int hostapd_global_init(struct hapd_interfaces *interfaces,
			       const char *entropy_file)
{
	int i;

	os_memset(&global, 0, sizeof(global));

	hostapd_logger_register_cb(hostapd_logger_cb);

	if (eap_server_register_methods()) {
		wpa_printf(MSG_ERROR, "Failed to register EAP methods");
		return -1;
	}

	if (eloop_init()) {
		wpa_printf(MSG_ERROR, "Failed to initialize event loop");
		return -1;
	}
	interfaces->eloop_initialized = 1;

	random_init(entropy_file);

#ifndef CONFIG_NATIVE_WINDOWS
	eloop_register_signal(SIGHUP, handle_reload, interfaces);
	eloop_register_signal(SIGUSR1, handle_dump_state, interfaces);
#endif /* CONFIG_NATIVE_WINDOWS */
	eloop_register_signal_terminate(handle_term, interfaces);

#ifndef CONFIG_NATIVE_WINDOWS
	openlog("hostapd", 0, LOG_DAEMON);
#endif /* CONFIG_NATIVE_WINDOWS */

	for (i = 0; wpa_drivers[i]; i++)
		global.drv_count++;
	if (global.drv_count == 0) {
		wpa_printf(MSG_ERROR, "No drivers enabled");
		return -1;
	}
	global.drv_priv = os_calloc(global.drv_count, sizeof(void *));
	if (global.drv_priv == NULL)
		return -1;

	return 0;
}


static void hostapd_global_deinit(const char *pid_file, int eloop_initialized)
{
	int i;

	for (i = 0; wpa_drivers[i] && global.drv_priv; i++) {
		if (!global.drv_priv[i])
			continue;
		wpa_drivers[i]->global_deinit(global.drv_priv[i]);
	}
	os_free(global.drv_priv);
	global.drv_priv = NULL;

#ifdef EAP_SERVER_TNC
	tncs_global_deinit();
#endif /* EAP_SERVER_TNC */

	random_deinit();

	if (eloop_initialized)
		eloop_destroy();

#ifndef CONFIG_NATIVE_WINDOWS
	closelog();
#endif /* CONFIG_NATIVE_WINDOWS */

	eap_server_unregister_methods();

	os_daemonize_terminate(pid_file);
}


static int hostapd_global_run(struct hapd_interfaces *ifaces, int daemonize,
			      const char *pid_file)
{
#ifdef EAP_SERVER_TNC
	int tnc = 0;
	size_t i, k;

	for (i = 0; !tnc && i < ifaces->count; i++) {
		for (k = 0; k < ifaces->iface[i]->num_bss; k++) {
			if (ifaces->iface[i]->bss[0]->conf->tnc) {
				tnc++;
				break;
			}
		}
	}

	if (tnc && tncs_global_init() < 0) {
		wpa_printf(MSG_ERROR, "Failed to initialize TNCS");
		return -1;
	}
#endif /* EAP_SERVER_TNC */

	if (daemonize) {
		if (os_daemonize(pid_file)) {
			wpa_printf(MSG_ERROR, "daemon: %s", strerror(errno));
			return -1;
		}
		if (eloop_sock_requeue()) {
			wpa_printf(MSG_ERROR, "eloop_sock_requeue: %s",
				   strerror(errno));
			return -1;
		}
	}

	eloop_run();

	return 0;
}


static void show_version(void)
{
	fprintf(stderr,
		"hostapd v" VERSION_STR "\n"
		"User space daemon for IEEE 802.11 AP management,\n"
		"IEEE 802.1X/WPA/WPA2/EAP/RADIUS Authenticator\n"
		"Copyright (c) 2002-2017, Jouni Malinen <j@w1.fi> "
		"and contributors\n");
}


static void usage(void)
{
	show_version();
	fprintf(stderr,
		"\n"
		"usage: hostapd [-hdBKtv] [-P <PID file>] [-e <entropy file>] "
		"\\\n"
		"         [-g <global ctrl_iface>] [-G <group>]\\\n"
		"         [-i <comma-separated list of interface names>]\\\n"
		"         <configuration file(s)>\n"
		"\n"
		"options:\n"
		"   -h   show this usage\n"
		"   -d   show more debug messages (-dd for even more)\n"
		"   -c   use only one config file\n"
		"   -B   run daemon in the background\n"
		"   -e   entropy file\n"
		"   -g   global control interface path\n"
		"   -G   group for control interfaces\n"
		"   -P   PID file\n"
		"   -K   include key data in debug messages\n"
#ifdef CONFIG_DEBUG_FILE
		"   -f   log output to debug file instead of stdout\n"
#endif /* CONFIG_DEBUG_FILE */
#ifdef CONFIG_DEBUG_LINUX_TRACING
		"   -T   record to Linux tracing in addition to logging\n"
		"        (records all messages regardless of debug verbosity)\n"
#endif /* CONFIG_DEBUG_LINUX_TRACING */
		"   -i   list of interface names to use\n"
#ifdef CONFIG_DEBUG_SYSLOG
		"   -s   log output to syslog instead of stdout\n"
#endif /* CONFIG_DEBUG_SYSLOG */
		"   -S   start all the interfaces synchronously\n"
		"   -t   include timestamps in some debug messages\n"
		"   -v   show hostapd version\n");

	exit(1);
}


static const char * hostapd_msg_ifname_cb(void *ctx)
{
	struct hostapd_data *hapd = ctx;
	if (hapd && hapd->conf)
		return hapd->conf->iface;
	return NULL;
}


static int hostapd_get_global_ctrl_iface(struct hapd_interfaces *interfaces,
					 const char *path)
{
#ifndef CONFIG_CTRL_IFACE_UDP
	char *pos;
#endif /* !CONFIG_CTRL_IFACE_UDP */

	os_free(interfaces->global_iface_path);
	interfaces->global_iface_path = os_strdup(path);
	if (interfaces->global_iface_path == NULL)
		return -1;

#ifndef CONFIG_CTRL_IFACE_UDP
	pos = os_strrchr(interfaces->global_iface_path, '/');
	if (pos == NULL) {
		wpa_printf(MSG_ERROR, "No '/' in the global control interface "
			   "file");
		os_free(interfaces->global_iface_path);
		interfaces->global_iface_path = NULL;
		return -1;
	}

	*pos = '\0';
	interfaces->global_iface_name = pos + 1;
#endif /* !CONFIG_CTRL_IFACE_UDP */

	return 0;
}


static int hostapd_get_ctrl_iface_group(struct hapd_interfaces *interfaces,
					const char *group)
{
#ifndef CONFIG_NATIVE_WINDOWS
	struct group *grp;
	grp = getgrnam(group);
	if (grp == NULL) {
		wpa_printf(MSG_ERROR, "Unknown group '%s'", group);
		return -1;
	}
	interfaces->ctrl_iface_group = grp->gr_gid;
#endif /* CONFIG_NATIVE_WINDOWS */
	return 0;
}
#ifdef CONFIG_WPS
static int gen_uuid(const char *txt_addr)
{
	u8 addr[ETH_ALEN];
	u8 uuid[UUID_LEN];
	char buf[100];

	if (hwaddr_aton(txt_addr, addr) < 0)
		return -1;

	uuid_gen_mac_addr(addr, uuid);
	if (uuid_bin2str(uuid, buf, sizeof(buf)) < 0)
		return -1;

	printf("%s\n", buf);

	return 0;
}
#endif /* CONFIG_WPS */


#ifndef HOSTAPD_CLEANUP_INTERVAL
#define HOSTAPD_CLEANUP_INTERVAL 10
#endif /* HOSTAPD_CLEANUP_INTERVAL */

static int hostapd_periodic_call(struct hostapd_iface *iface, void *ctx)
{
	hostapd_periodic_iface(iface);
	return 0;
}


/* Periodic cleanup tasks */
static void hostapd_periodic(void *eloop_ctx, void *timeout_ctx)
{
	struct hapd_interfaces *interfaces = eloop_ctx;

	eloop_register_timeout(HOSTAPD_CLEANUP_INTERVAL, 0,
			       hostapd_periodic, interfaces, NULL);
	hostapd_for_each_interface(interfaces, hostapd_periodic_call, NULL);
}


int main(int argc, char *argv[])
{
	struct hapd_interfaces interfaces;
	int ret = 1;
	int vlan = 0;
	size_t i, j;
	int c, debug = 0, daemonize = 0;
	char *pid_file = NULL;
	const char *log_file = NULL;
	const char *entropy_file = NULL;
	char **bss_config = NULL, **tmp_bss;
	size_t num_bss_configs = 0;
#ifdef CONFIG_DEBUG_LINUX_TRACING
	int enable_trace_dbg = 0;
#endif /* CONFIG_DEBUG_LINUX_TRACING */
	int start_ifaces_in_sync = 0;
	char **if_names = NULL;
	size_t if_names_size = 0;
	char **if_mab_names = NULL;
	size_t if_mab_names_size = 0;
	char **if_pvlan_names = NULL;
	size_t if_pvlan_names_size = 0;
	char **if_voice_names = NULL;
	size_t if_voice_names_size = 0;
	char *config_fname = NULL;
	struct hostapd_config *conf = NULL;
	const char aclcommand[] = "/usr/cumulus/bin/cl-acltool";
	const char aclargs[] = "-i";

	if (os_program_init())
		return -1;

	os_memset(&interfaces, 0, sizeof(interfaces));
	interfaces.reload_config = hostapd_reload_config;
	interfaces.config_read_cb = hostapd_config_read;
	interfaces.for_each_interface = hostapd_for_each_interface;
	interfaces.ctrl_iface_init = hostapd_ctrl_iface_init;
	interfaces.ctrl_iface_deinit = hostapd_ctrl_iface_deinit;
	interfaces.driver_init = hostapd_driver_init;
	interfaces.global_iface_path = NULL;
	interfaces.global_iface_name = NULL;
	interfaces.global_ctrl_sock = -1;
	dl_list_init(&interfaces.global_ctrl_dst);
#ifdef CONFIG_ETH_P_OUI
	dl_list_init(&interfaces.eth_p_oui);
#endif /* CONFIG_ETH_P_OUI */
#ifdef CONFIG_DPP
	hostapd_dpp_init_global(&interfaces);
#endif /* CONFIG_DPP */

	for (;;) {
		c = getopt(argc, argv, "b:Bc:de:f:hi:KP:sSTtu:vg:G:");
		if (c < 0)
			break;
		switch (c) {
		case 'h':
			usage();
			break;
		case 'd':
			debug++;
			if (wpa_debug_level > 0)
				wpa_debug_level--;
			break;
		case 'B':
			daemonize++;
			break;
		case 'e':
			entropy_file = optarg;
			break;
		case 'f':
			log_file = optarg;
			break;
		case 'K':
			wpa_debug_show_keys++;
			break;
		case 'P':
			os_free(pid_file);
			pid_file = os_rel2abs_path(optarg);
			break;
		case 't':
			wpa_debug_timestamp++;
			break;
#ifdef CONFIG_DEBUG_LINUX_TRACING
		case 'T':
			enable_trace_dbg = 1;
			break;
#endif /* CONFIG_DEBUG_LINUX_TRACING */
		case 'v':
			show_version();
			exit(1);
			break;
		case 'g':
			if (hostapd_get_global_ctrl_iface(&interfaces, optarg))
				return -1;
			break;
		case 'G':
			if (hostapd_get_ctrl_iface_group(&interfaces, optarg))
				return -1;
			break;
		case 'b':
			tmp_bss = os_realloc_array(bss_config,
						   num_bss_configs + 1,
						   sizeof(char *));
			if (tmp_bss == NULL)
				goto out;
			bss_config = tmp_bss;
			bss_config[num_bss_configs++] = optarg;
			break;
#ifdef CONFIG_DEBUG_SYSLOG
		case 's':
			wpa_debug_syslog = 1;
			break;
#endif /* CONFIG_DEBUG_SYSLOG */
		case 'S':
			start_ifaces_in_sync = 1;
			break;
#ifdef CONFIG_WPS
		case 'u':
			return gen_uuid(optarg);
#endif /* CONFIG_WPS */
		case 'c':
			/* there is one config and it has everything */
			config_fname = optarg;
			break;
		case 'i':
			if (hostapd_get_interface_names(&if_names,
							&if_names_size, optarg))
				goto out;
			break;
		default:
			usage();
			break;
		}
	}

	if (optind == argc && interfaces.global_iface_path == NULL &&
	    num_bss_configs == 0 && config_fname == NULL)
		usage();

	if (config_fname && if_names_size) {
		/* we should not have both -i and -c used */
		wpa_printf(MSG_ERROR, "Option -i cannot be used with -c");
		goto out;
	}

	wpa_msg_register_ifname_cb(hostapd_msg_ifname_cb);

	if (log_file)
		wpa_debug_open_file(log_file);
	else
		wpa_debug_setup_stdout();
#ifdef CONFIG_DEBUG_SYSLOG
	if (wpa_debug_syslog)
		wpa_debug_open_syslog();
#endif /* CONFIG_DEBUG_SYSLOG */
#ifdef CONFIG_DEBUG_LINUX_TRACING
	if (enable_trace_dbg) {
		int tret = wpa_debug_open_linux_tracing();
		if (tret) {
			wpa_printf(MSG_ERROR, "Failed to enable trace logging");
			return -1;
		}
	}
#endif /* CONFIG_DEBUG_LINUX_TRACING */

	if (if_names && if_names_size) {
		/* the -i was used to specify interfaces so that
		 * we already have if_names
		 */
		interfaces.count = if_names_size;
		/* the last arg must be the one config file */
		config_fname = argv[optind];
	} else if (config_fname) {
		/*
		 * We have one config file where all interfaces are specified
		 * with interfaces=swp1,swp2,swp3. If this is the case, we need
		 * to get the interfaces first. The interfaces all share the
		 * same config attributes. For the wired driver, this is often
		 * the case.
		 */
		if (hostapd_get_interface_names_config(&if_names, &if_names_size,
						       HOSTAPD_BASE_IFACE_CONFIG,
						       config_fname))
			goto out;
		interfaces.count = if_names_size;
		ret = hostapd_get_interface_names_config(&if_mab_names, &if_mab_names_size,
						         HOSTAPD_MAB_IFACE_CONFIG,
							 config_fname);
		ret = hostapd_get_interface_names_config(&if_pvlan_names, &if_pvlan_names_size,
				                         HOSTAPD_PVLAN_IFACE_CONFIG,
							 config_fname);
		ret = hostapd_get_interface_names_config(&if_voice_names, &if_voice_names_size,
				                         HOSTAPD_VOICE_IFACE_CONFIG,
							 config_fname);
		/* grab the config for radius das purposes */
		conf = interfaces.config_read_cb(config_fname);
	} else {
		/* we have the case were each interface has a config file */
		interfaces.count = argc - optind;
	}

	if (interfaces.count || num_bss_configs) {
		interfaces.iface = os_calloc(interfaces.count + num_bss_configs,
					     sizeof(struct hostapd_iface *));
		if (interfaces.iface == NULL) {
			wpa_printf(MSG_ERROR, "malloc failed");
			return -1;
		}
	}

	if (hostapd_global_init(&interfaces, entropy_file)) {
		wpa_printf(MSG_ERROR, "Failed to initialize global context");
		return -1;
	}

	eloop_register_timeout(HOSTAPD_CLEANUP_INTERVAL, 0,
			       hostapd_periodic, &interfaces, NULL);

	if (fst_global_init()) {
		wpa_printf(MSG_ERROR,
			   "Failed to initialize global FST context");
		goto out;
	}

#if defined(CONFIG_FST) && defined(CONFIG_CTRL_IFACE)
	if (!fst_global_add_ctrl(fst_ctrl_cli))
		wpa_printf(MSG_WARNING, "Failed to add CLI FST ctrl");
#endif /* CONFIG_FST && CONFIG_CTRL_IFACE */

	/* Let the rest of the drivers know we are starting */
	interfaces.start_in_progress = TRUE;
	interfaces.stop_in_progress = FALSE;

	/* Allocate and parse configuration for full interface files */
	for (i = 0; i < interfaces.count; i++) {
		char *if_name = NULL;

		if (i < if_names_size)
			if_name = if_names[i];

		if (config_fname == NULL)
			config_fname = argv[optind + i];

		interfaces.iface[i] = hostapd_interface_init(&interfaces,
							     if_name,
							     config_fname,
							     debug);
		if (!interfaces.iface[i]) {
			wpa_printf(MSG_ERROR, "Failed to initialize interface");
			goto out;
		}

		/* some interfaces need to be tagged for MAB */
		for (j = 0; j < if_mab_names_size; j++) {
			if (interfaces.iface[i]->bss[0] &&
					os_strcmp(if_name, if_mab_names[j]) == 0) {
				interfaces.iface[i]->bss[0]->conf->mab_enabled = TRUE;
				break;
			}
		}
		/* some interfaces need to be tagged for to use parking vlans */
		for (j = 0; j < if_pvlan_names_size; j++) {
			if (interfaces.iface[i]->bss[0] &&
					os_strcmp(if_name, if_pvlan_names[j]) == 0) {
				interfaces.iface[i]->bss[0]->conf->pvlan_enabled = TRUE;
				break;
			}
		}
		/* some interfaces need to be tagged for as voice interfaces */
		for (j = 0; j < if_voice_names_size; j++) {
			if (interfaces.iface[i]->bss[0] && (hostapd_compare_voice_ifnames(if_name, if_voice_names[j]) == 0)) {
				interfaces.iface[i]->bss[0]->conf->voice_enabled = TRUE;
				/* check for a VLAN (usually for voice) */
				vlan = hostapd_get_interface_vlan(if_voice_names[j]);
				if (vlan)
					interfaces.iface[i]->bss[0]->conf->voice_vlan_id = vlan;
				break;
			}
		}

		if (start_ifaces_in_sync)
			interfaces.iface[i]->need_to_start_in_sync = 1;
	}

	/* Allocate and parse configuration for per-BSS files */
	for (i = 0; i < num_bss_configs; i++) {
		struct hostapd_iface *iface;
		char *fname;

		wpa_printf(MSG_INFO, "BSS config: %s", bss_config[i]);
		fname = os_strchr(bss_config[i], ':');
		if (fname == NULL) {
			wpa_printf(MSG_ERROR,
				   "Invalid BSS config identifier '%s'",
				   bss_config[i]);
			goto out;
		}
		*fname++ = '\0';
		iface = hostapd_interface_init_bss(&interfaces, bss_config[i],
						   fname, debug);
		if (iface == NULL)
			goto out;
		for (j = 0; j < interfaces.count; j++) {
			if (interfaces.iface[j] == iface)
				break;
		}
		if (j == interfaces.count) {
			struct hostapd_iface **tmp;
			tmp = os_realloc_array(interfaces.iface,
					       interfaces.count + 1,
					       sizeof(struct hostapd_iface *));
			if (tmp == NULL) {
				hostapd_interface_deinit_free(iface);
				goto out;
			}
			interfaces.iface = tmp;
			interfaces.iface[interfaces.count++] = iface;
		}
	}

	/*
	 * Before enabling interfaces, make sure to flush any existing
	 * ACL rules files that 802.1X created.
	 */
	hostapd_handle_acl_pvid_removal(NULL);

	/*
	 * Enable configured interfaces. Depending on channel configuration,
	 * this may complete full initialization before returning or use a
	 * callback mechanism to complete setup in case of operations like HT
	 * co-ex scans, ACS, or DFS are needed to determine channel parameters.
	 * In such case, the interface will be enabled from eloop context within
	 * hostapd_global_run().
	 */
	interfaces.terminate_on_error = interfaces.count;
	for (i = 0; i < interfaces.count; i++) {
		if (hostapd_driver_init(interfaces.iface[i]))
			goto out;
#ifdef CONFIG_MBO
		for (j = 0; j < interfaces.iface[i]->num_bss; j++) {
			struct hostapd_data *hapd = interfaces.iface[i]->bss[j];

			if (hapd && (hapd->conf->oce & OCE_STA_CFON) &&
			    (interfaces.iface[i]->drv_flags &
			     WPA_DRIVER_FLAGS_OCE_STA_CFON))
				hapd->enable_oce = OCE_STA_CFON;

			if (hapd && (hapd->conf->oce & OCE_AP) &&
			    (interfaces.iface[i]->drv_flags &
			     WPA_DRIVER_FLAGS_OCE_STA_CFON)) {
				/* TODO: Need to add OCE-AP support */
				wpa_printf(MSG_ERROR,
					   "OCE-AP feature is not yet supported");
			}
		}
#endif /* CONFIG_MBO */
		if (hostapd_setup_interface(interfaces.iface[i]))
			goto out;
	}

	hostapd_global_ctrl_iface_init(&interfaces);

	/* now startup the RADIUS DAS server if needed.  We only have one
	   server for hostapd not a server per interface as before. */
	if (conf && conf->bss[0] && (conf->bss[0]->radius_das_port != 0)) {
		struct radius_das_conf das_conf;
		struct hostapd_bss_config *bss_conf = conf->bss[0];

		os_memset(&das_conf, 0, sizeof(das_conf));
		das_conf.port = bss_conf->radius_das_port;

		/* we can have multiple DAS clients sending CoA commands */
		for (i = 0; i < bss_conf->radius_das_client_count; i++) {
			das_conf.radius_das_clients[i].shared_secret =
				bss_conf->radius_das_clients[i].shared_secret;
			das_conf.radius_das_clients[i].shared_secret_len =
				bss_conf->radius_das_clients[i].shared_secret_len;
			das_conf.radius_das_clients[i].client_addr =
				bss_conf->radius_das_clients[i].client_addr;
		}
		os_strlcpy(das_conf.vrf, bss_conf->radius_das_client_vrf,
				sizeof(das_conf.vrf));

		das_conf.time_window = bss_conf->radius_das_time_window;
		das_conf.nas_identifier = bss_conf->nas_identifier;
		das_conf.radius_das_client_count = bss_conf->radius_das_client_count;
		das_conf.require_event_timestamp =
			bss_conf->radius_das_require_event_timestamp;
		das_conf.require_message_authenticator =
			bss_conf->radius_das_require_message_authenticator;
		if (interfaces.count > 0)
			das_conf.ctx = &interfaces;
		das_conf.disconnect = hostapd_das_global_disconnect;
		das_conf.handle_coa = hostapd_das_global_handle_coa;
		interfaces.radius_das = radius_das_init(&das_conf);
		if (interfaces.radius_das == NULL) {
			wpa_printf(MSG_ERROR, "RADIUS DAS initialization "
				   "failed.");
			return -1;
		}


	}

	/* The driver should have written the rules for each interface
	 * so we only call this once at the end.
	 */
	wired_acl_signal(NULL, NULL);

	/* Let the rest of the drivers know we are done starting */
	interfaces.start_in_progress = FALSE;
	sd_notify(0, "READY=1"); /*  notify systemd that we are ready */

	if (hostapd_global_run(&interfaces, daemonize, pid_file)) {
		wpa_printf(MSG_ERROR, "Failed to start eloop");
		goto out;
	}

	ret = 0;

 out:
	/* Let the rest of the drivers know we are coming down */
	interfaces.stop_in_progress = TRUE;

	hostapd_global_ctrl_iface_deinit(&interfaces);
	/* Deinitialize all interfaces */
	for (i = 0; i < interfaces.count; i++) {
		if (!interfaces.iface[i])
			continue;
		interfaces.iface[i]->driver_ap_teardown =
			!!(interfaces.iface[i]->drv_flags &
			   WPA_DRIVER_FLAGS_AP_TEARDOWN_SUPPORT);
		hostapd_interface_deinit_free(interfaces.iface[i]);
	}
	os_free(interfaces.iface);

#ifdef CONFIG_DPP
	hostapd_dpp_deinit_global(&interfaces);
#endif /* CONFIG_DPP */

	if (interfaces.eloop_initialized)
		eloop_cancel_timeout(hostapd_periodic, &interfaces, NULL);
	hostapd_global_deinit(pid_file, interfaces.eloop_initialized);
	radius_das_deinit(interfaces.radius_das);
	os_free(pid_file);

	if (os_exec(aclcommand, aclargs, 1)) {
		wpa_printf(MSG_ERROR, "WIRED %s: could not run cl-acltool",
				__FUNCTION__);
	}
	wpa_debug_close_syslog();
	if (log_file)
		wpa_debug_close_file();
	wpa_debug_close_linux_tracing();

	os_free(bss_config);

	for (i = 0; i < if_names_size; i++)
		os_free(if_names[i]);
	os_free(if_names);

	fst_global_deinit();

	os_program_deinit();

	/*
	 * When SIGTERM or SIGINT is caught, we will need to remove any existing
	 * ACL rules. systemd may restart hostapd and reinstall default rules.
	 */
	/* hostapd_handle_acl_pvid_removal(NULL); */
	/* The driver should have removed all rules for each interface
	 * so we only call this once at the end.
	 */
	sd_notify(0, "STOPPING=1"); /*  notify systemd that we are stopping */

	return ret;
}
